#include <limits.h> // for ULONG_MAX
#include <time.h>
#include "secret.h"
#include "atlib.h"
#include "ntplib.h"

#define VERBOSE // Comment this for a totally silent operation

#define NTP_PKT_SIZE 48
#define RHO_MAX 0.001
#define RHORES_MIN 0.000015
#define EPSILON_MAX 200
//#define NTP_SERVER "time.nist.gov" //Non risponde!
//#define NTP_SERVER "pool.ntp.org"    // typ 75ms (strato 2)
//#define NTP_SERVER "ntp1.inrim.it"   // typ 100ms (strato 1)
//#define NTP_SERVER "at.pool.ntp.org" // typ 100 ms (strato 2)
#define NTP_SERVER "ntp.metas.ch"    // typ 93ms (strato 1)

typedef struct sync {
  unsigned long t;        // the local time of the synchronization event
  NTPtime tNTP;           // the proper time of the synchronization event (estimated)
  NTPtime Delta;          // the difference between the two (tNTP - t)
  unsigned long epsilon;  // the margin of error of the estimate (uncertainty)
};

int freeRam () {
  extern int __heap_start, *__brkval; 
  int v; 
  return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval); 
}


// ESP esp1(9,8,19200,10);   // Costruttore: RX,TX,baudrate,CH_PD : Uno+quartz
ESP esp1(9,10,19200,8); // : Pro mini + ceramic
sync S_prec;              // Data from the last synchronization event
sync S_now;               // Data from the current synchronization event
boolean noHistory=true;   // True if there is a valid record of the last synchronization event
float rho=0.0;            // The last computed drift value (rho)

/* This function called at powerUp of the ESP */
//boolean pup() { true;}

void setup() {
#ifdef VERBOSE
  Serial.begin(57600);    // Impostazione baudrate USB Arduino
  Serial.println("\n===\nInizio test");
#endif //VERBOSE
//  esp1.powerUp(30,{});
}

/* Given a value in msecs of the local clock as the first parameter, and the result of the last synchronization
 * event (second and third parameter), updates the fourth parameter with an estimate of the proper
 * time corresponding with the local clock. In practice, it is used to obtain proper timestamps for
 * local events
  */
void VirtualClock(unsigned long t, float rho, sync *S, NTPtime *vcNTP) {
  NTPtime offsetNTP;
  unsigned long offset; // Assumes  offset is positive (always after some day in the 70's)
  long tmp = ((float)(t - S->t) * rho); // may be positive or negative
//  offset = (t - S->t) + tmp; // problems when t-S->t over 2^31, it's converted to long
  if ( tmp > 0 ) {
     offset = (t - S->t) + (unsigned long)tmp;
  } else {
     offset = (t - S->t) - (unsigned long)(-tmp);
  }
  ms2NTP(&offsetNTP,offset);
  NTPsum(&(S->tNTP),&offsetNTP,vcNTP); 
  return;
}

/* Print a readable proper timestamp associated to a local clock value tx */ 
void printTime(unsigned long tx) {
  NTPtime vcNTP;
  char timestamp[40];
  int ms;
  VirtualClock(tx,rho,&S_prec,&vcNTP);
//  printNTP(&(vcNTP),"Virtual time",3);
  setTime(vcNTP.sec);
  ms=round(((float)(vcNTP.frac>>18)/(1<<14))*1000);
 // digital clock display of the time
  sprintf(timestamp," %d/%d/%d %d:%02d:%02d.%03d",
    day(),month(),year(),hour(),minute(),second(),ms);
  Serial.print(timestamp);
}

/*
 * Encodes a 64-bit timestamp in the NTP packet as an array of two long integers: the 
 * integral part, converted in Unix time, and the fractional part in fixed Q32 format.
 * The p paramente is the index of the first octet in the packet ntppkt, and the result
 * is returned in the ntptime array
 */
void gettime(int p, byte ntppkt[], unsigned long ntptime[]) {
  for ( int i=0; i<8; i++ ) {
    ntptime[i/4] <<= 8; ntptime[i/4] |= ntppkt[p+i];
  }
  ntptime[0] = ntptime[0] - 2208988800L;
  return;
}

/* Performs a synchronization event with the NTP server indicated in the first parameter,
 * and updates the sync structure pointed by the second parameter with the data (see above).
 * The function prepares the NTP request that is sent (UDP) to the server. When the response
 * is received the function evaluates the roundtrip time, and computes all the parameters needed
 * to fill the synchronization report S: the remote  time (proper) at the time of sending the
 * request, the estimated local time for send event on the server, the offset between the two,
 * the error range. If an error is found during the process, the function returns false,
 * otherwise true.
 */
int synchronize(char server[], sync *S) {
  unsigned long t0,t1; //send,receive
  unsigned long ntptime[2];
  unsigned long loctime[2];
  byte ntppkt[NTP_PKT_SIZE+1];
  char cmd[80];
//  Serial.println(freeRam());
  ntppkt[0] = 0x1B; // li=00 ver=011 mode=011 (client)
  ntppkt[1] = 0x0F; // poll=0 stratum=15
  for (int i=2;i<NTP_PKT_SIZE;i++) {ntppkt[i] = 0X0;}
  sprintf(cmd,"AT+CIPSTART=\"UDP\",\"%s\",123",server);
  if ( ! esp1.atcmd(cmd,2) ) {
    #ifdef VERBOSE
    Serial.println("Errore connessione");
#endif //VERBOSE
    return 1;
  }
  sprintf(cmd,"AT+CIPSEND=%d",NTP_PKT_SIZE);
  esp1.atcmd(cmd,5);
// NTP roundtrip begins ==========================================
  t0=millis();
  esp1.sendpkt(ntppkt, NTP_PKT_SIZE);
  if ( esp1.readpkt(ntppkt, NTP_PKT_SIZE, 5) != NTP_PKT_SIZE ) {
#ifdef VERBOSE
    Serial.println("Errore pacchetto ricevuto");
#endif //VERBOSE
    esp1.atcmd("AT+CIPCLOSE",5);
    return 1;
  }
  t1=millis();
// NTP roundtrip ends ========================================
  esp1.atcmd("AT+CIPCLOSE",5);       // TCP: Connessione HTTP
#ifdef VERBOSE
  Serial.print("Server=");
  Serial.println(server);
  Serial.print("Stratum=");
  Serial.println(ntppkt[1],DEC);
  Serial.print("Tsnd=");
#endif //VERBOSE
  gettime(40,ntppkt,ntptime);
#ifdef VERBOSE
  Serial.print (ntptime[0]); Serial.print(" + ");
  Serial.println(((float)(ntptime[1]>>20)/(1<<12)),6);
#endif //VERBOSE
  S->tNTP.sec=ntptime[0];
  S->tNTP.frac=ntptime[1];
//* Calcola roundtrip e tempo compensato tc sul server */
  S->epsilon=(t1-t0)/2;
  S->t = t0 + S->epsilon;

#ifdef VERBOSE
  Serial.print("t_0 (msec)=");
  Serial.println(S->t);
  Serial.print("epsilon (msec)=");
  Serial.println(S->epsilon);
  Serial.print("Offset (NTP)=");
#endif //VERBOSE
/* Converte tc in notazione NTP */
  loctime[0]=(S->t)/1000;
  loctime[1]=((((S->t)%1000)*(1<<12))/1000)<<20;
  if (ntptime[1]>loctime[1]) {
    S->Delta.frac=ntptime[1]-loctime[1];
    S->Delta.sec=ntptime[0]-loctime[0];
  } else {
    S->Delta.frac=~(loctime[1]-ntptime[1]);
    S->Delta.sec=ntptime[0]-loctime[0]-1;
  }
#ifdef VERBOSE
  Serial.print(S->Delta.sec); Serial.print(" + ");
  Serial.println(((float)(S->Delta.frac>>20)/(1<<12)),3);
#endif //VERBOSE
  return 0;
}

void loop() {
  int sign;
  long num;
  float rho_res=RHO_MAX;// default drift values
  NTPtime vcNTP, dNTP;
  unsigned long tx=millis()+60000;
  
  Serial.println("===");
//  Serial.println(freeRam());
//  Call the synchronize function. In case of problems, wait and repeat
/* Power-up the ESP-01 */
  esp1.powerUp(10,{});
  while ( synchronize(NTP_SERVER,&S_now) != 0 ) {
    esp1.powerDown();
#ifdef VERBOSE  
    Serial.println("Fail: retry in 20 secs");
#endif //VERBOSE
    delay(20000);
    esp1.powerUp(10,{});  
  }
  esp1.powerDown();
//  
//  // Qui il calcolo del timeout (attenzione agli overflow) DAFARE
// If there is a previous record, compute changes.
  if (! noHistory) { // otherwise use default values
    VirtualClock(S_now.t,rho,&S_prec,&vcNTP);
#ifdef VERBOSE
    printNTP(&(vcNTP),"Virtual time at t_0 (with old rho)",3);    // Test only: check if near to Tsnd
#endif //VERBOSE
// Compute the rho, using the current and the last synchronization
    sign=NTPdiff(&S_now.Delta,&S_prec.Delta,&dNTP);   // NTP difference needed to avoid overflow
    num=sign*NTP2ms(&dNTP);
#ifdef VERBOSE
    Serial.print("dD (msec)=");
    Serial.println(num);
#endif //VERBOSE
    rho=num/(float)(S_now.t-S_prec.t);            // rho=(d0-d1)/(t0-t1)
#ifdef VERBOSE
    Serial.print("rho (drift ppm)=");
    Serial.println(rho*1000000.0);
#endif //VERBOSE
// Compute the error range of the rho    
    rho_res=(S_now.epsilon+S_prec.epsilon)/(float)(S_now.t-S_prec.t);       // rho'=(eps0+eps1)/(t0-t1)  
    if (rho_res < RHORES_MIN) { rho_res = RHORES_MIN; }
#ifdef VERBOSE
    Serial.print("rho' (residual ppm)=");
    Serial.println(rho_res*1000000.0);
#endif VERBOSE
  }
/* Compute the delay for the next synchronization event, based on estimate rho.
 * To avoid delays longer than the rollover period, we compute a lower threshold value
 * for the residual rho: if the residual rho is lower than that threshold, the delay
 * would be longer than the rollover period. If that is the case the timeout is set to
 * the current event time, thus avoiding a delay longer than the rollover. The rollover
 * is around 50 days (49d 17h 2m 24s).
 */
  unsigned long diffEpsilon=EPSILON_MAX-S_now.epsilon;
  float rho_min=diffEpsilon/ULONG_MAX;  //check rollover
  if ( rho_res<rho_min ) {  
    tx=S_now.t;                       // rollover case, wait full rollover period
  } else {
    tx=S_now.t+diffEpsilon/rho_res;
  }
#ifdef VERBOSE
  Serial.print("Next sync (msec)=");
  Serial.println(tx);
#endif //VERBOSE
/* Manage the next loop state */
  S_prec=S_now;
  noHistory=false;
//  // Da provare con "micros()" for delays that extend after the rollover
//  if ( millis() > tx ) {
//    while (millis() > tx) {};
//  }

  unsigned long tC=millis();
  while (millis() < tx) {        // while waiting of the timeout
#ifdef VERBOSE
    unsigned long tN = millis();
    if ( (tN - tC) >= 30000 ) {  // every 30 seconds
      Serial.print("  ");        // print the local time (millis)
      Serial.print(tN);
      Serial.print(" -> ");
      printTime(tN);               // and the correspsonding time (
//      while ( synchronize(NTP_SERVER,&S_now) != 0 ) {
//        Serial.println("Fail: retry in 20 secs");
//        delay(20000);
//      }
//      long offset = NTP2ms(&S_now.epsilon);
//      Serial.print("\t");
//      Serial.print(exp_offset);
//      Serial.print("\t");
//      Serial.print(offset);
      Serial.println();
      tC=tN;
      }
#endif //VERBOSE
  };
}
